#include <unistd.h>

#include "mbedtls/net.h"
#include "mbedtls/ssl.h"
#include "mbedtls/entropy.h"
#include "mbedtls/ctr_drbg.h"
#ifdef DEBUG
#include "mbedtls/debug.h"
#endif

#include "common.h"

mbedtls_net_context         server_fd;
mbedtls_entropy_context     entropy;
mbedtls_ctr_drbg_context    ctr_drbg;
mbedtls_ssl_context         ctx;
mbedtls_ssl_config          conf;
mbedtls_x509_crt            cacert;

// Buffer used for read/write tests
unsigned char rw_buf[BUFFER_SIZE];

// Function calls 'mbedtls_ssl_handshake' and 'mbedtls_ssl_close_notify'
#define HANDLE_OP_RW_NEEDS(op, ctx)                                                     \
do {                                                                                    \
    while  ((ret = (op)((ctx)))) {                                                      \
        if( ret != MBEDTLS_ERR_SSL_WANT_READ && ret != MBEDTLS_ERR_SSL_WANT_WRITE ) {   \
            ERR("Critical error when performing " #op);                                 \
        }                                                                               \
    }                                                                                   \
} while (0);


#ifdef DEBUG
static void my_debug( void *ctx, int level,
                      const char *file, int line, const char *str ) {
    ((void) level);
    fprintf( (FILE *) ctx, "%s:%04d: %s", file, line, str );
    fflush(  (FILE *) ctx  );
}
#endif

void init() {
    int ret;

    mbedtls_net_init( &server_fd );
    mbedtls_ssl_init( &ctx );
    mbedtls_ssl_config_init( &conf );
    mbedtls_x509_crt_init( &cacert );
    mbedtls_ctr_drbg_init( &ctr_drbg );
    mbedtls_entropy_init( &entropy );

    mbedtls_ssl_conf_rng( &conf, mbedtls_ctr_drbg_random, &ctr_drbg );
    ret = mbedtls_ctr_drbg_seed( &ctr_drbg, mbedtls_entropy_func, &entropy,
                               (const unsigned char *) CLIENT,
                               strlen( CLIENT ) );
    if (ret) {
        ERR( "mbedtls_ctr_drbg_seed");
    }

    ret = mbedtls_ssl_config_defaults(&conf,
                    MBEDTLS_SSL_IS_CLIENT,
                    MBEDTLS_SSL_TRANSPORT_STREAM, // TLS not DTLS
                    MBEDTLS_SSL_PRESET_DEFAULT );
    if (ret) {
        ERR("mbedtls_ssl_config_defaults");
    }

    mbedtls_ssl_conf_authmode( &conf, MBEDTLS_SSL_VERIFY_REQUIRED );
    if (ret) {
        ERR( "mbedtls_ssl_conf_authmode");
    }

    ret = mbedtls_x509_crt_parse_file(&cacert, CACERT);
    if (ret) {
        ERR("mbedtls_x509_crt_parse");
    }

    mbedtls_ssl_conf_ca_chain( &conf, &cacert, NULL );
    //mbedtls_ssl_conf_ciphersuites( &conf, AllowedCiphers);
#ifdef DEBUG
    mbedtls_ssl_conf_dbg( &conf, my_debug, stdout );
    mbedtls_debug_set_threshold(4);
#endif
     // end of 'conf' object setup

    ret = mbedtls_ssl_setup(&ctx, &conf);
    if (ret) {
        ERR("mbedtls_ssl_setup");
    }
}

void cleanup(void) {
    mbedtls_x509_crt_free(&cacert);
    mbedtls_net_free( &server_fd );
    mbedtls_ssl_free( &ctx );
    mbedtls_ssl_config_free( &conf );
    mbedtls_ctr_drbg_free( &ctr_drbg );
    mbedtls_entropy_free( &entropy );
}

static void setup_client_ctx(void)
{

    if (mbedtls_ssl_set_hostname(&ctx, CERT_HOSTNAME) ) {
        ERR("mbedtls_ssl_set_hostname");
    }
    mbedtls_ssl_set_bio(&ctx, &server_fd, mbedtls_net_send, mbedtls_net_recv, NULL );
}

void test_Handshake(size_t handshake_nb) {
    int ret;
    init();

    DBG("Trying to connect");
    server_fd.fd = connect_once();
    setup_client_ctx();

    for (int i=0; i<handshake_nb; i++) {

        DBG("SSL handshake");
        HANDLE_OP_RW_NEEDS(mbedtls_ssl_handshake, &ctx);

        DBG("Send 'close notify'");
        HANDLE_OP_RW_NEEDS(mbedtls_ssl_close_notify, &ctx);

        DBG("SSL reset");
        mbedtls_ssl_session_reset(&ctx);
    }

    close(server_fd.fd);
    cleanup();
}

void test_Write() {
    int ret, err, nread =0;
    init();

    DBG("Trying to connect");
    server_fd.fd = connect_once();
    setup_client_ctx();

    DBG("SSL handshake");
    HANDLE_OP_RW_NEEDS(mbedtls_ssl_handshake, &ctx);

    fprintf(stderr, "%s\n", mbedtls_ssl_get_version(&ctx));
    fprintf(stderr, "%s\n", mbedtls_ssl_get_ciphersuite(&ctx));

    fill_buffer_from_file(rw_buf, sizeof(rw_buf));
    for (size_t i=0; i<RW_ITERATIONS; i++) {
        for(nread=0; nread<sizeof(rw_buf); nread+=err) {
            err = mbedtls_ssl_write(&ctx, rw_buf+nread, sizeof(rw_buf) - nread);
            if (err == MBEDTLS_ERR_SSL_WANT_READ && err == MBEDTLS_ERR_SSL_WANT_WRITE) {
                // From spec: when this function returns MBEDTLS_ERR_SSL_WANT_WRITE/READ,
                // it must be called later with the same arguments, until
                // it returns a positive value.
                err = 0;
                continue;
            }

            if(err <= 0) {
                ERR("mbedtls_ssl_write");
            }
        }
    }

    DBG("Send 'close notify'");
    HANDLE_OP_RW_NEEDS(mbedtls_ssl_close_notify, &ctx);

    close(server_fd.fd);
    cleanup();

}

void test_Read() {

    int ret, err, nread = 0;

    init();

    DBG("Trying to connect");
    server_fd.fd = connect_once();
    setup_client_ctx();

    DBG("SSL handshake");
    HANDLE_OP_RW_NEEDS(mbedtls_ssl_handshake, &ctx);

    if (mbedtls_ssl_write( &ctx, (unsigned char*)READ_REQUEST, sizeof(READ_REQUEST) ) <= 0) {
        ERR("mbedtls_ssl_write");
    }

    do {
#ifdef DEBUG
        fwrite(rw_buf,1, nread, stdout);
#endif
        for(nread=0; nread<sizeof(rw_buf); nread+=err) {
            err = mbedtls_ssl_read(&ctx, rw_buf+nread, sizeof(rw_buf) - nread);
            if (err == MBEDTLS_ERR_SSL_WANT_READ && err == MBEDTLS_ERR_SSL_WANT_WRITE) {
                // I assume as for mbedtls_ssl_write, the read function will handle
                // MBEDTLS_ERR_SSL_WANT_WRITE/READ return codes
                err = 0;
                continue;
            }

            if(!err) break;
        }
    } while(err>0);

    if (err != MBEDTLS_ERR_SSL_PEER_CLOSE_NOTIFY) {
        ERR("mbedtls_ssl_read");
    }

    DBG("Send 'close notify'");
    HANDLE_OP_RW_NEEDS(mbedtls_ssl_close_notify, &ctx);

    close(server_fd.fd);
    cleanup();
}

int main(int argc, char* argv[]) {

    if (argc < 2) {
        goto usage;
    }

    if (!strncmp("test_Handshake", argv[1], strlen("test_Handshake"))) {
        test_Handshake(HANDHAKE_REPS);
    } else if (!strncmp("test_Read", argv[1], strlen("test_Read"))) {
        test_Read();
    } else if (!strncmp("test_Write", argv[1], strlen("test_Write"))) {
        test_Write();
    } else {
        printf("Unknown test");
        goto usage;
    }

exit:
    return 0;

usage:
    ERR("\n\nUsage: %s test_name \nOptions for 'test_name':\n\t"
        "test_Handshake\n\ttest_Write\n\ttest_Read\n", argv[0]);
    goto exit;
}